This is the release version of EAR replication.

User's Guide
===

I. Download Hadoop-20
---
Hadoop-20 can be downloaded from [here](https://github.com/facebookarchive/hadoop-20).  
Download and extract it in one of your hosts.
If the above link is not working, you can download from [homepage of this project] (http://ansrlab.cse.cuhk.edu.hk/software/ear).

To simplify our description, in this instruction, we refer to the hadoop-20/ directory as hadoop directory.  

Call the following command to update the environment variable.
> export HADOOP_HOME=*absolute address of hadoop directory*   
> export PATH=$PATH:$HADOOP_HOME/bin   

It is a good practice to include the above two lines in your ~/.bashrc file.

Add the following line to $HADOOP_HOME/conf/hadoop-env.sh
> export JAVA_HOME=*YOUR JAVA_HOME*

II. Installation and Configuration
---
0. Run `bash install.sh` to install EAR.

1. Configure masters and slaves  
  To start with, you need to make sure in your cluster every machine is
  public-key accessible from other machines.  
  Input the IP address/hostname of master and slaves in file conf/masters and
  conf/slaves.  

2. Create cluster topology definition script.
  Hadoop uses a topology definition script for rack awareness.  
  The script takes IP address or host name of a machine as input and returns
  which rack the machine belongs to.  Take a look at $HADOOP_HOME/rackAware.sh
  to get a feeling and change it according to your cluster topology.

  To use EAR, we assume you have enough racks in your cluster to support a
  valid placement.  For example, you cannot require that a 14-block stripe to be
  placed in a 10-rack cluster with c=1.

3. Configure conf/core-site.xml conf/hdfs-site.xml conf/raid.xml   
  according to the instructions in the xml files.

4. Finishing: 
  Copy the hadoop directory to every machine in the same location.

IV. Walk through
---
Using default settings, a file is first written using two-way replication and
later encoded using a (6,4) RS code.  Suppose that you have six DataNodes, say
node1 to node6.  Configure them to be in six different racks, rack1 to rack6,
respectively.   
Suppose the directory to do raid is called raidTest. 

Let us first format and start HDFS
> hadoop namenode -format   
> start-dfs.sh   

Go to node1 and prepare a 256MB (4-block) file, name it testFile.   
Write the file in to HDFS.  
> hadoop dfs -copyFromLocal testFile raidTest/testFile

The four blocks form a stripe, each block has a replica placed in node1.  Thus
/rack1 is the core rack for this stripe.

Now you can grep the pattern "2 datanodes" from the log file of namenode in
$HADOOP_HOME/logs and figure out the locations of the four blocks you just
copied into HDFS.  
After deleting one redundant replica of each block, the remaining replicas are
located in different racks.

To do encoding, start MapReduce and RaidNode.
> start-mapred.sh  
> start-raidnode.sh  

Wait until the pattern "Job Complete" appears in the log file of raidnode.
In master node, call "grep Choosing logs/\*", and you can find out the
jobtracker assigns the encoding task to /rack1/node1 (node in core-rack).

Developer's Guide
===

IMPORTANT NOTE: EAR replication will be installed as a patch to Hadoop-20,
after installation.  If you want to revert, please back up the original source
codes before you install EAR.

II. Understanding the Source Code. 
---

There are mainly three components of the codes we modified.

1. The Placement module
   (src/contrib/raid/src/java/org/apache/hadoop/hdfs/server/namenode/): We
   start from placing the replicated blocks.  
      - EARLayoutGen.java (newly created) generates layout for a stripe of data
        blocks.  
      - EARBlockPlacementPolicy.java (newly created) returns block locations
        for a specific block.
      - PreEncodingStripeStore.java (newly created) manages the pre-encoding
        stripe store for future encoding job.

2. The RaidNode module (src/contrib/raid/src/java/org/apache/hadoop/raid/):  
      - DistRaid.java (updated) reorganizes the stripes for encoding and
        realizes locality of encoding tasks.
      - RaidNode.java (updated) loads the stripe information from pre-encoding
        stripe store.

3. The MapReduce module (src/mapred/org/apache/hadoop/mapred/):  
      - JobConf.java (updated) to add flag to identify encoding job.  
      - EARSequenceFileRecordReader.java (newly created) for interpreting input
        file of encoding job. 

Change Log
===

- Ver.1.0.1 (August 19th, 2015): Bug fix.
- Ver.1.0.0: First release.



